@@ -1,53 +1,35 @@
 /*
- * Copyright (c) 2016 Digital Bazaar, Inc. All rights reserved.
+ * Copyright (c) 2016-2021 Digital Bazaar, Inc. All rights reserved.
  */
 'use strict';
 
 const IdentifierIssuer = require('./IdentifierIssuer');
 const MessageDigest = require('./MessageDigest');
-const Permutator = require('./Permutator');
+const Permuter = require('./Permuter');
 const NQuads = require('./NQuads');
-const util = require('./util');
-
-const POSITIONS = {subject: 's', object: 'o', graph: 'g'};
 
 module.exports = class URDNA2015Sync {
   constructor() {
     this.name = 'URDNA2015';
-    this.blankNodeInfo = {};
-    this.hashToBlankNodes = {};
+    this.blankNodeInfo = new Map();
     this.canonicalIssuer = new IdentifierIssuer('_:c14n');
     this.hashAlgorithm = 'sha256';
-    this.quads;
+    this.quads = null;
   }
 
   // 4.4) Normalization Algorithm
   main(dataset) {
-    const self = this;
-    self.quads = dataset;
+    this.quads = dataset;
 
     // 1) Create the normalization state.
-
-    // Note: Optimize by generating non-normalized blank node map concurrently.
-    const nonNormalized = {};
-
     // 2) For every quad in input dataset:
     for(const quad of dataset) {
       // 2.1) For each blank node that occurs in the quad, add a reference
       // to the quad using the blank node identifier in the blank node to
       // quads map, creating a new entry if necessary.
-      self.forEachComponent(quad, component => {
-        if(component.termType !== 'BlankNode') {
-          return;
-        }
-        const id = component.value;
-        if(id in self.blankNodeInfo) {
-          self.blankNodeInfo[id].quads.push(quad);
-        } else {
-          nonNormalized[id] = true;
-          self.blankNodeInfo[id] = {quads: [quad]};
-        }
-      });
+      this._addBlankNodeQuadInfo({quad, component: quad.subject});
+      this._addBlankNodeQuadInfo({quad, component: quad.object});
+      this._addBlankNodeQuadInfo({quad, component: quad.graph});
     }
 
     // 3) Create a list of non-normalized blank node identifiers
@@ -55,80 +37,64 @@ module.exports = class URDNA2015Sync {
     // blank node to quads map.
     // Note: We use a map here and it was generated during step 2.
 
-    // 4) Initialize simple, a boolean flag, to true.
-    let simple = true;
+    // 4) `simple` flag is skipped -- loop is optimized away. This optimization
+    // is permitted because there was a typo in the hash first degree quads
+    // algorithm in the URDNA2015 spec that was implemented widely making it
+    // such that it could not be fixed; the result was that the loop only
+    // needs to be run once and the first degree quad hashes will never change.
+    // 5.1-5.2 are skipped; first degree quad hashes are generated just once
+    // for all non-normalized blank nodes.
 
-    // 5) While simple is true, issue canonical identifiers for blank nodes:
-    while(simple) {
-      // 5.1) Set simple to false.
-      simple = false;
+    // 5.3) For each blank node identifier identifier in non-normalized
+    // identifiers:
+    const hashToBlankNodes = new Map();
+    const nonNormalized = [...this.blankNodeInfo.keys()];
+    for(const id of nonNormalized) {
+      // steps 5.3.1 and 5.3.2:
+      this._hashAndTrackBlankNode({id, hashToBlankNodes});
+    }
 
-      // 5.2) Clear hash to blank nodes map.
-      self.hashToBlankNodes = {};
-
-      // 5.3) For each blank node identifier identifier in non-normalized
-      // identifiers:
-      for(const id in nonNormalized) {
-        // 5.3.1) Create a hash, hash, according to the Hash First Degree
-        // Quads algorithm.
-        const hash = self.hashFirstDegreeQuads(id);
-
-        // 5.3.2) Add hash and identifier to hash to blank nodes map,
-        // creating a new entry if necessary.
-        if(hash in self.hashToBlankNodes) {
-          self.hashToBlankNodes[hash].push(id);
-        } else {
-          self.hashToBlankNodes[hash] = [id];
-        }
+    // 5.4) For each hash to identifier list mapping in hash to blank
+    // nodes map, lexicographically-sorted by hash:
+    const hashes = [...hashToBlankNodes.keys()].sort();
+    // optimize away second sort, gather non-unique hashes in order as we go
+    const nonUnique = [];
+    for(const hash of hashes) {
+      // 5.4.1) If the length of identifier list is greater than 1,
+      // continue to the next mapping.
+      const idList = hashToBlankNodes.get(hash);
+      if(idList.length > 1) {
+        nonUnique.push(idList);
+        continue;
       }
 
-      // 5.4) For each hash to identifier list mapping in hash to blank
-      // nodes map, lexicographically-sorted by hash:
-      const hashes = Object.keys(self.hashToBlankNodes).sort();
-      for(let i = 0; i < hashes.length; ++i) {
-        // 5.4.1) If the length of identifier list is greater than 1,
-        // continue to the next mapping.
-        const hash = hashes[i];
-        const idList = self.hashToBlankNodes[hash];
-        if(idList.length > 1) {
-          continue;
-        }
+      // 5.4.2) Use the Issue Identifier algorithm, passing canonical
+      // issuer and the single blank node identifier in identifier
+      // list, identifier, to issue a canonical replacement identifier
+      // for identifier.
+      const id = idList[0];
+      this.canonicalIssuer.getId(id);
 
-        // 5.4.2) Use the Issue Identifier algorithm, passing canonical
-        // issuer and the single blank node identifier in identifier
-        // list, identifier, to issue a canonical replacement identifier
-        // for identifier.
-        // TODO: consider changing `getId` to `issue`
-        const id = idList[0];
-        self.canonicalIssuer.getId(id);
-
-        // 5.4.3) Remove identifier from non-normalized identifiers.
-        delete nonNormalized[id];
-
-        // 5.4.4) Remove hash from the hash to blank nodes map.
-        delete self.hashToBlankNodes[hash];
-
-        // 5.4.5) Set simple to true.
-        simple = true;
-      }
+      // Note: These steps are skipped, optimized away since the loop
+      // only needs to be run once.
+      // 5.4.3) Remove identifier from non-normalized identifiers.
+      // 5.4.4) Remove hash from the hash to blank nodes map.
+      // 5.4.5) Set simple to true.
     }
 
     // 6) For each hash to identifier list mapping in hash to blank nodes map,
     // lexicographically-sorted by hash:
-    const hashes = Object.keys(self.hashToBlankNodes).sort();
-    for(let i = 0; i < hashes.length; ++i) {
+    // Note: sort optimized away, use `nonUnique`.
+    for(const idList of nonUnique) {
       // 6.1) Create hash path list where each item will be a result of
       // running the Hash N-Degree Quads algorithm.
       const hashPathList = [];
 
       // 6.2) For each blank node identifier identifier in identifier list:
-      const hash = hashes[i];
-      const idList = self.hashToBlankNodes[hash];
-      for(let j = 0; j < idList.length; ++j) {
+      for(const id of idList) {
         // 6.2.1) If a canonical identifier has already been issued for
         // identifier, continue to the next identifier.
-        const id = idList[j];
-        if(self.canonicalIssuer.hasId(id)) {
+        if(this.canonicalIssuer.hasId(id)) {
           continue;
         }
 
@@ -143,24 +109,22 @@ module.exports = class URDNA2015Sync {
 
         // 6.2.4) Run the Hash N-Degree Quads algorithm, passing
         // temporary issuer, and append the result to the hash path list.
-        const result = self.hashNDegreeQuads(id, issuer);
+        const result = this.hashNDegreeQuads(id, issuer);
         hashPathList.push(result);
       }
 
       // 6.3) For each result in the hash path list,
       // lexicographically-sorted by the hash in result:
-      // TODO: use `String.localeCompare`?
-      hashPathList.sort((a, b) =>
-        (a.hash < b.hash) ? -1 : ((a.hash > b.hash) ? 1 : 0));
-      for(let j = 0; j < hashPathList.length; ++j) {
+      hashPathList.sort(_stringHashCompare);
+      for(const result of hashPathList) {
         // 6.3.1) For each blank node identifier, existing identifier,
         // that was issued a temporary identifier by identifier issuer
         // in result, issue a canonical identifier, in the same order,
         // using the Issue Identifier algorithm, passing canonical
         // issuer and existing identifier.
-        const result = hashPathList[j];
-        for(const existing in result.issuer.existing) {
-          self.canonicalIssuer.getId(existing);
+        const oldIds = result.issuer.getOldIds();
+        for(const id of oldIds) {
+          this.canonicalIssuer.getId(id);
         }
       }
     }
@@ -172,20 +136,17 @@ module.exports = class URDNA2015Sync {
 
     // 7) For each quad, quad, in input dataset:
     const normalized = [];
-    for(let i = 0; i < self.quads.length; ++i) {
+    for(const quad of this.quads) {
       // 7.1) Create a copy, quad copy, of quad and replace any existing
       // blank node identifiers using the canonical identifiers
       // previously issued by canonical issuer.
-      // Note: We optimize away the copy here.
-      const quad = self.quads[i];
-      self.forEachComponent(quad, component => {
-        if(component.termType === 'BlankNode' &&
-          !component.value.startsWith(self.canonicalIssuer.prefix)) {
-          component.value = self.canonicalIssuer.getId(component.value);
-        }
-      });
+      // Note: We optimize with shallow copies here.
+      const q = {...quad};
+      q.subject = this._useCanonicalId({component: q.subject});
+      q.object = this._useCanonicalId({component: q.object});
+      q.graph = this._useCanonicalId({component: q.graph});
       // 7.2) Add quad copy to the normalized dataset.
-      normalized.push(NQuads.serializeQuad(quad));
+      normalized.push(NQuads.serializeQuad(q));
     }
 
     // sort normalized output
@@ -197,38 +158,34 @@ module.exports = class URDNA2015Sync {
 
   // 4.6) Hash First Degree Quads
   hashFirstDegreeQuads(id) {
-    const self = this;
-
-    // return cached hash
-    const info = self.blankNodeInfo[id];
-    if('hash' in info) {
-      return info.hash;
-    }
-
     // 1) Initialize nquads to an empty list. It will be used to store quads in
     // N-Quads format.
     const nquads = [];
 
     // 2) Get the list of quads `quads` associated with the reference blank node
     // identifier in the blank node to quads map.
+    const info = this.blankNodeInfo.get(id);
     const quads = info.quads;
 
     // 3) For each quad `quad` in `quads`:
-    for(let i = 0; i < quads.length; ++i) {
-      const quad = quads[i];
-
+    for(const quad of quads) {
       // 3.1) Serialize the quad in N-Quads format with the following special
       // rule:
 
       // 3.1.1) If any component in quad is an blank node, then serialize it
       // using a special identifier as follows:
-      const copy = {predicate: quad.predicate};
-      self.forEachComponent(quad, (component, key) => {
-        // 3.1.2) If the blank node's existing blank node identifier matches
-        // the reference blank node identifier then use the blank node
-        // identifier _:a, otherwise, use the blank node identifier _:z.
-        copy[key] = self.modifyFirstDegreeComponent(id, component, key);
-      });
+      const copy = {
+        subject: null, predicate: quad.predicate, object: null, graph: null
+      };
+      // 3.1.2) If the blank node's existing blank node identifier matches
+      // the reference blank node identifier then use the blank node
+      // identifier _:a, otherwise, use the blank node identifier _:z.
+      copy.subject = this.modifyFirstDegreeComponent(
+        id, quad.subject, 'subject');
+      copy.object = this.modifyFirstDegreeComponent(
+        id, quad.object, 'object');
+      copy.graph = this.modifyFirstDegreeComponent(
+        id, quad.graph, 'graph');
       nquads.push(NQuads.serializeQuad(copy));
     }
 
@@ -237,41 +194,38 @@ module.exports = class URDNA2015Sync {
 
     // 5) Return the hash that results from passing the sorted, joined nquads
     // through the hash algorithm.
-    const md = new MessageDigest(self.hashAlgorithm);
-    for(let i = 0; i < nquads.length; ++i) {
-      md.update(nquads[i]);
+    const md = new MessageDigest(this.hashAlgorithm);
+    for(const nquad of nquads) {
+      md.update(nquad);
     }
-    // TODO: represent as byte buffer instead to cut memory usage in half
     info.hash = md.digest();
     return info.hash;
   }
 
   // 4.7) Hash Related Blank Node
   hashRelatedBlankNode(related, quad, issuer, position) {
-    const self = this;
-
     // 1) Set the identifier to use for related, preferring first the canonical
     // identifier for related if issued, second the identifier issued by issuer
     // if issued, and last, if necessary, the result of the Hash First Degree
     // Quads algorithm, passing related.
     let id;
-    if(self.canonicalIssuer.hasId(related)) {
-      id = self.canonicalIssuer.getId(related);
+    if(this.canonicalIssuer.hasId(related)) {
+      id = this.canonicalIssuer.getId(related);
     } else if(issuer.hasId(related)) {
       id = issuer.getId(related);
     } else {
-      id = self.hashFirstDegreeQuads(related);
+      id = this.blankNodeInfo.get(related).hash;
     }
 
     // 2) Initialize a string input to the value of position.
     // Note: We use a hash object instead.
-    const md = new MessageDigest(self.hashAlgorithm);
+    const md = new MessageDigest(this.hashAlgorithm);
     md.update(position);
 
     // 3) If position is not g, append <, the value of the predicate in quad,
     // and > to input.
     if(position !== 'g') {
-      md.update(self.getRelatedPredicate(quad));
+      md.update(this.getRelatedPredicate(quad));
     }
 
     // 4) Append identifier to input.
@@ -279,29 +233,25 @@ module.exports = class URDNA2015Sync {
 
     // 5) Return the hash that results from passing input through the hash
     // algorithm.
-    // TODO: represent as byte buffer instead to cut memory usage in half
     return md.digest();
   }
 
   // 4.8) Hash N-Degree Quads
   hashNDegreeQuads(id, issuer) {
-    const self = this;
-
     // 1) Create a hash to related blank nodes map for storing hashes that
     // identify related blank nodes.
     // Note: 2) and 3) handled within `createHashToRelated`
-    const md = new MessageDigest(self.hashAlgorithm);
-    const hashToRelated = self.createHashToRelated(id, issuer);
+    const md = new MessageDigest(this.hashAlgorithm);
+    const hashToRelated = this.createHashToRelated(id, issuer);
 
     // 4) Create an empty string, data to hash.
     // Note: We created a hash object `md` above instead.
 
     // 5) For each related hash to blank node list mapping in hash to related
     // blank nodes map, sorted lexicographically by related hash:
-    const hashes = Object.keys(hashToRelated).sort();
-    for(let i = 0; i < hashes.length; ++i) {
+    const hashes = [...hashToRelated.keys()].sort();
+    for(const hash of hashes) {
       // 5.1) Append the related hash to the data to hash.
-      const hash = hashes[i];
       md.update(hash);
 
       // 5.2) Create a string chosen path.
@@ -311,9 +261,9 @@ module.exports = class URDNA2015Sync {
       let chosenIssuer;
 
       // 5.4) For each permutation of blank node list:
-      const permutator = new Permutator(hashToRelated[hash]);
-      while(permutator.hasNext()) {
-        const permutation = permutator.next();
+      const permuter = new Permuter(hashToRelated.get(hash));
+      while(permuter.hasNext()) {
+        const permutation = permuter.next();
 
         // 5.4.1) Create a copy of issuer, issuer copy.
         let issuerCopy = issuer.clone();
@@ -327,12 +277,11 @@ module.exports = class URDNA2015Sync {
 
         // 5.4.4) For each related in permutation:
         let nextPermutation = false;
-        for(let j = 0; j < permutation.length; ++j) {
+        for(const related of permutation) {
           // 5.4.4.1) If a canonical identifier has been issued for
           // related, append it to path.
-          const related = permutation[j];
-          if(self.canonicalIssuer.hasId(related)) {
-            path += self.canonicalIssuer.getId(related);
+          if(this.canonicalIssuer.hasId(related)) {
+            path += this.canonicalIssuer.getId(related);
           } else {
             // 5.4.4.2) Otherwise:
             // 5.4.4.2.1) If issuer copy has not issued an identifier for
@@ -362,19 +311,18 @@ module.exports = class URDNA2015Sync {
         }
 
         // 5.4.5) For each related in recursion list:
-        for(let j = 0; j < recursionList.length; ++j) {
+        for(const related of recursionList) {
           // 5.4.5.1) Set result to the result of recursively executing
           // the Hash N-Degree Quads algorithm, passing related for
           // identifier and issuer copy for path identifier issuer.
-          const related = recursionList[j];
-          const result = self.hashNDegreeQuads(related, issuerCopy);
+          const result = this.hashNDegreeQuads(related, issuerCopy);
 
           // 5.4.5.2) Use the Issue Identifier algorithm, passing issuer
           // copy and related and append the result to path.
           path += issuerCopy.getId(related);
 
           // 5.4.5.3) Append <, the hash in result, and > to path.
-          path += '<' + result.hash + '>';
+          path += `<${result.hash}>`;
 
           // 5.4.5.4) Set issuer copy to the identifier issuer in
           // result.
@@ -422,71 +370,119 @@ module.exports = class URDNA2015Sync {
     if(component.termType !== 'BlankNode') {
       return component;
     }
-    component = util.clone(component);
-    component.value = (component.value === id ? '_:a' : '_:z');
-    return component;
+    /* Note: A mistake in the URDNA2015 spec that made its way into
+    implementations (and therefore must stay to avoid interop breakage)
+    resulted in an assigned canonical ID, if available for
+    `component.value`, not being used in place of `_:a`/`_:z`, so
+    we don't use it here. */
+    return {
+      termType: 'BlankNode',
+      value: component.value === id ? '_:a' : '_:z'
+    };
   }
 
   // helper for getting a related predicate
   getRelatedPredicate(quad) {
-    return '<' + quad.predicate.value + '>';
+    return `<${quad.predicate.value}>`;
   }
 
   // helper for creating hash to related blank nodes map
   createHashToRelated(id, issuer) {
-    const self = this;
-
     // 1) Create a hash to related blank nodes map for storing hashes that
     // identify related blank nodes.
-    const hashToRelated = {};
+    const hashToRelated = new Map();
 
     // 2) Get a reference, quads, to the list of quads in the blank node to
     // quads map for the key identifier.
-    const quads = self.blankNodeInfo[id].quads;
+    const quads = this.blankNodeInfo.get(id).quads;
 
     // 3) For each quad in quads:
-    for(let i = 0; i < quads.length; ++i) {
+    for(const quad of quads) {
       // 3.1) For each component in quad, if component is the subject, object,
-      // and graph name and it is a blank node that is not identified by
+      // or graph name and it is a blank node that is not identified by
       // identifier:
-      const quad = quads[i];
-      for(const key in quad) {
-        const component = quad[key];
-        if(key === 'predicate' ||
-          !(component.termType === 'BlankNode' && component.value !== id)) {
-          continue;
-        }
-        // 3.1.1) Set hash to the result of the Hash Related Blank Node
-        // algorithm, passing the blank node identifier for component as
-        // related, quad, path identifier issuer as issuer, and position as
-        // either s, o, or g based on whether component is a subject, object,
-        // graph name, respectively.
-        const related = component.value;
-        const position = POSITIONS[key];
-        const hash = self.hashRelatedBlankNode(related, quad, issuer, position);
-
-        // 3.1.2) Add a mapping of hash to the blank node identifier for
-        // component to hash to related blank nodes map, adding an entry as
-        // necessary.
-        if(hash in hashToRelated) {
-          hashToRelated[hash].push(related);
-        } else {
-          hashToRelated[hash] = [related];
-        }
-      }
+      // steps 3.1.1 and 3.1.2 occur in helpers:
+      this._addRelatedBlankNodeHash({
+        quad, component: quad.subject, position: 's',
+        id, issuer, hashToRelated
+      });
+      this._addRelatedBlankNodeHash({
+        quad, component: quad.object, position: 'o',
+        id, issuer, hashToRelated
+      });
+      this._addRelatedBlankNodeHash({
+        quad, component: quad.graph, position: 'g',
+        id, issuer, hashToRelated
+      });
     }
 
     return hashToRelated;
   }
 
-  // helper that iterates over quad components (skips predicate)
-  forEachComponent(quad, op) {
-    for(const key in quad) {
-      // skip `predicate`
-      if(key === 'predicate') {
-        continue;
-      }
-      op(quad[key], key, quad);
+  _hashAndTrackBlankNode({id, hashToBlankNodes}) {
+    // 5.3.1) Create a hash, hash, according to the Hash First Degree
+    // Quads algorithm.
+    const hash = this.hashFirstDegreeQuads(id);
+
+    // 5.3.2) Add hash and identifier to hash to blank nodes map,
+    // creating a new entry if necessary.
+    const idList = hashToBlankNodes.get(hash);
+    if(!idList) {
+      hashToBlankNodes.set(hash, [id]);
+    } else {
+      idList.push(id);
     }
   }
+
+  _addBlankNodeQuadInfo({quad, component}) {
+    if(component.termType !== 'BlankNode') {
+      return;
+    }
+    const id = component.value;
+    const info = this.blankNodeInfo.get(id);
+    if(info) {
+      info.quads.add(quad);
+    } else {
+      this.blankNodeInfo.set(id, {quads: new Set([quad]), hash: null});
+    }
+  }
+
+  _addRelatedBlankNodeHash(
+    {quad, component, position, id, issuer, hashToRelated}) {
+    if(!(component.termType === 'BlankNode' && component.value !== id)) {
+      return;
+    }
+    // 3.1.1) Set hash to the result of the Hash Related Blank Node
+    // algorithm, passing the blank node identifier for component as
+    // related, quad, path identifier issuer as issuer, and position as
+    // either s, o, or g based on whether component is a subject, object,
+    // graph name, respectively.
+    const related = component.value;
+    const hash = this.hashRelatedBlankNode(related, quad, issuer, position);
+
+    // 3.1.2) Add a mapping of hash to the blank node identifier for
+    // component to hash to related blank nodes map, adding an entry as
+    // necessary.
+    const entries = hashToRelated.get(hash);
+    if(entries) {
+      entries.push(related);
+    } else {
+      hashToRelated.set(hash, [related]);
+    }
+  }
+
+  _useCanonicalId({component}) {
+    if(component.termType === 'BlankNode' &&
+      !component.value.startsWith(this.canonicalIssuer.prefix)) {
+      return {
+        termType: 'BlankNode',
+        value: this.canonicalIssuer.getId(component.value)
+      };
+    }
+    return component;
+  }
 };
+
+function _stringHashCompare(a, b) {
+  return a.hash < b.hash ? -1 : a.hash > b.hash ? 1 : 0;
+}
