@@ -1,14 +1,13 @@
 /*
- * Copyright (c) 2016-2017 Digital Bazaar, Inc. All rights reserved.
+ * Copyright (c) 2016-2021 Digital Bazaar, Inc. All rights reserved.
  */
 'use strict';
 
 const URDNA2015 = require('./URDNA2015');
-const util = require('./util');
 
 module.exports = class URDNA2012 extends URDNA2015 {
-  constructor(options) {
-    super(options);
+  constructor() {
+    super();
     this.name = 'URGNA2012';
     this.hashAlgorithm = 'sha1';
   }
@@ -18,13 +17,16 @@ module.exports = class URDNA2012 extends URDNA2015 {
     if(component.termType !== 'BlankNode') {
       return component;
     }
-    component = util.clone(component);
-    if(key === 'name') {
-      component.value = '_:g';
-    } else {
-      component.value = (component.value === id ? '_:a' : '_:z');
+    if(key === 'graph') {
+      return {
+        termType: 'BlankNode',
+        value: '_:g'
+      };
     }
-    return component;
+    return {
+      termType: 'BlankNode',
+      value: (component.value === id ? '_:a' : '_:z')
+    };
   }
 
   // helper for getting a related predicate
@@ -33,19 +35,18 @@ module.exports = class URDNA2012 extends URDNA2015 {
   }
 
   // helper for creating hash to related blank nodes map
-  createHashToRelated(id, issuer, callback) {
-    const self = this;
-
+  async createHashToRelated(id, issuer) {
     // 1) Create a hash to related blank nodes map for storing hashes that
     // identify related blank nodes.
-    const hashToRelated = {};
+    const hashToRelated = new Map();
 
     // 2) Get a reference, quads, to the list of quads in the blank node to
     // quads map for the key identifier.
-    const quads = self.blankNodeInfo[id].quads;
+    const quads = this.blankNodeInfo.get(id).quads;
 
     // 3) For each quad in quads:
-    self.forEach(quads, (quad, idx, callback) => {
+    let i = 0;
+    for(const quad of quads) {
       // 3.1) If the quad's subject is a blank node that does not match
       // identifier, set hash to the result of the Hash Related Blank Node
       // algorithm, passing the blank node identifier for subject as related,
@@ -65,23 +66,25 @@ module.exports = class URDNA2012 extends URDNA2015 {
         position = 'r';
       } else {
         // 3.3) Otherwise, continue to the next quad.
-        return callback();
+        continue;
+      }
+      // Note: batch hashing related blank nodes 100 at a time
+      if(++i % 100 === 0) {
+        await this._yield();
       }
       // 3.4) Add a mapping of hash to the blank node identifier for the
       // component that matched (subject or object) to hash to related blank
       // nodes map, adding an entry as necessary.
-      self.hashRelatedBlankNode(
-        related, quad, issuer, position, (err, hash) => {
-          if(err) {
-            return callback(err);
-          }
-          if(hash in hashToRelated) {
-            hashToRelated[hash].push(related);
-          } else {
-            hashToRelated[hash] = [related];
-          }
-          callback();
-        });
-    }, err => callback(err, hashToRelated));
+      const hash = await this.hashRelatedBlankNode(
+        related, quad, issuer, position);
+      const entries = hashToRelated.get(hash);
+      if(entries) {
+        entries.push(related);
+      } else {
+        hashToRelated.set(hash, [related]);
+      }
+    }
+
+    return hashToRelated;
   }
 };
