@@ -3,7 +3,7 @@
  * This library works in the browser and node.js.
  *
  * BSD 3-Clause License
- * Copyright (c) 2016-2017 Digital Bazaar, Inc.
+ * Copyright (c) 2016-2021 Digital Bazaar, Inc.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -34,7 +34,6 @@
  */
 'use strict';
 
-const util = require('./util');
 const URDNA2015 = require('./URDNA2015');
 const URGNA2012 = require('./URGNA2012');
 const URDNA2015Sync = require('./URDNA2015Sync');
@@ -75,57 +74,42 @@ api._rdfCanonizeNative = function(api) {
  *          algorithm the canonicalization algorithm to use, `URDNA2015` or
  *            `URGNA2012`.
  *          [useNative] use native implementation (default: false).
- * @param [callback(err, canonical)] called once the operation completes.
  *
  * @return a Promise that resolves to the canonicalized RDF Dataset.
  */
-api.canonize = util.callbackify(async function(dataset, options) {
-  let callback;
-  const promise = new Promise((resolve, reject) => {
-    callback = (err, canonical) => {
-      if(err) {
-        return reject(err);
-      }
-
-      /*if(options.format === 'application/n-quads') {
-        canonical = canonical.join('');
-      }
-      canonical = _parseNQuads(canonical.join(''));*/
-
-      resolve(canonical);
-    };
-  });
-
+api.canonize = async function(dataset, options) {
   // back-compat with legacy dataset
   if(!Array.isArray(dataset)) {
     dataset = api.NQuads.legacyDatasetToQuads(dataset);
   }
 
-  // TODO: convert algorithms to Promise-based async
   if(options.useNative) {
-    if(rdfCanonizeNative) {
-      rdfCanonizeNative.canonize(dataset, options, callback);
-    } else {
+    if(!rdfCanonizeNative) {
       throw new Error('rdf-canonize-native not available');
     }
-  } else {
-    if(options.algorithm === 'URDNA2015') {
-      new URDNA2015(options).main(dataset, callback);
-    } else if(options.algorithm === 'URGNA2012') {
-      new URGNA2012(options).main(dataset, callback);
-    } else if(!('algorithm' in options)) {
-      throw new Error('No RDF Dataset Canonicalization algorithm specified.');
-    } else {
-      throw new Error(
-        'Invalid RDF Dataset Canonicalization algorithm: ' + options.algorithm);
-    }
+    // TODO: convert native algorithm to Promise-based async
+    return new Promise((resolve, reject) =>
+      rdfCanonizeNative.canonize(dataset, options, (err, canonical) =>
+        err ? reject(err) : resolve(canonical)));
   }
 
-  return promise;
-});
+  if(options.algorithm === 'URDNA2015') {
+    return new URDNA2015(options).main(dataset);
+  }
+  if(options.algorithm === 'URGNA2012') {
+    return new URGNA2012(options).main(dataset);
+  }
+  if(!('algorithm' in options)) {
+    throw new Error('No RDF Dataset Canonicalization algorithm specified.');
+  }
+  throw new Error(
+    'Invalid RDF Dataset Canonicalization algorithm: ' + options.algorithm);
+};
 
 /**
- * Synchronously canonizes an RDF dataset.
+ * This method is no longer available in the public API, it is for testing
+ * only. It synchronously canonizes an RDF dataset and does not work in the
+ * browser.
  *
  * @param dataset the dataset to canonize.
  * @param options the options to use:
@@ -135,7 +119,7 @@ api.canonize = util.callbackify(async function(dataset, options) {
  *
  * @return the RDF dataset in canonical form.
  */
-api.canonizeSync = function(dataset, options) {
+api._canonizeSync = function(dataset, options) {
   // back-compat with legacy dataset
   if(!Array.isArray(dataset)) {
     dataset = api.NQuads.legacyDatasetToQuads(dataset);
@@ -149,7 +133,8 @@ api.canonizeSync = function(dataset, options) {
   }
   if(options.algorithm === 'URDNA2015') {
     return new URDNA2015Sync(options).main(dataset);
-  } else if(options.algorithm === 'URGNA2012') {
+  }
+  if(options.algorithm === 'URGNA2012') {
     return new URGNA2012Sync(options).main(dataset);
   }
   if(!('algorithm' in options)) {
